import {
  useState, useRef, useCallback, useEffect,
} from 'react';
import { fromEvent } from 'rxjs';
import {
  throttleTime,
} from 'rxjs/operators';

interface CursorPosition {
  x: number | null;
  y: number | null;
  clientX: number | null;
  clientY: number | null;
  screenX: number | null;
  screenY: number | null;
  isOver: boolean;
  isDown: boolean;
}

const initialState = {
  x: null,
  y: null,
  clientX: null,
  clientY: null,
  screenX: null,
  screenY: null,
  isOver: false,
  isDown: false,
};

const useCursorPosition = (): [CursorPosition, (element: HTMLElement | null) => void] => {
  const [state, setState] = useState<CursorPosition>(initialState);
  const [element, setElement] = useState<HTMLElement | null>(null);
  const entered = useRef(false);

  const onMoveEvent = useCallback((e: any): void => {
    if (element) {
      const {
        clientX, clientY, screenX, screenY, pageX, pageY,
      } = e;
      const rect = element.getBoundingClientRect();
      const offsetX = window.pageXOffset || window.scrollX || 0;
      const offsetY = window.pageYOffset || window.scrollY || 0;

      setState((prev: CursorPosition) => ({
        x: pageX - rect.left - offsetX,
        y: pageY - rect.top - offsetY,
        clientX,
        clientY,
        screenX,
        screenY,
        isOver: true,
        isDown: prev.isDown,
      }));
    }
  }, [element]);

  useEffect(() => {
    if (element) {
      let subscription: any = null;
      const onEnter = (): void => {
        entered.current = true;
      };

      const onLeave = (): void => {
        entered.current = false;
        setState(initialState);
      };

      const onDown = (e: MouseEvent): void => {
        entered.current = true;
        onMoveEvent(e);
        subscription = fromEvent(element, 'mousemove').pipe(throttleTime(35)).subscribe(onMoveEvent);
      };

      const onUp = (): void => {
        entered.current = false;
        setState(initialState);
        if (subscription) subscription.unsubscribe();
      };

      const addEvent = element.addEventListener.bind(element);
      addEvent('mouseenter', onEnter);
      addEvent('mouseleave', onLeave);
      addEvent('mousedown', onDown);
      addEvent('mouseup', onUp);
    }
  }, [element]);

  return [state, setElement];
};

export default useCursorPosition;
